unit uSudokuGrid;

interface
uses
  classes, sysutils, grids, windows, graphics, forms, controls, menus,
  uSudokuBase;

const
  CellStrongBorder = 2;
  CellThinBorder = 1;
  CellWidth = 50;
  CellHeight = 50;
  
type
  // Composant graphique permettant d'afficher la grille  l'cran et saisir les valeurs des cellules.
  TGrilleSudoku = class(TCustomGrid)
  private
    FSudokuGrid : TSudokuBase;  // La grille de Sudoku  dessiner.

    FBorderBrush : TBrush;        // Remplissage des bordures.
    FCellBrush : TBrush;          // Remplissage des cellules normales.
    FSelectedCellBrush : TBrush;  // Remplissage de la cellule slectionne.
    FReveleeFont : TFont;         // Font pour dessiner les cellules rvles.
    FNormalFont : TFont;          // Font pour dessiner les cellules joues par le joueur.
    FMarkFont : TFont;            // Font pour dessiner les valeurs candidates.

    FPopupMenu : TPopupMenu;      // Menu contextuel affich par la grille.
    FMarkRoot : TMenuItem;        // Racine du sous-menu pour dfinir les candidats.
    FCurrentCell : TGridCoord;    // Coordonnes de la grille slectionne.
    FAffichageLettre: boolean;    // Indique si les cellules sont dessines avec des chiffres ou des lettres.

    procedure SetSudokuGrid(const Value: TSudokuBase);
    procedure SetAffichageLettre(const Value: boolean);
  protected
    // Dessine la cellule <ACol, ARow>. Il s'agit bien videmment de la mthode la plus importante du
    // composant.
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect;
      AState: TGridDrawState); override;

    // Ragit au clic sur un bouton de la souris pour afficher le menu contextuel.
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;

    // Ragit au dplacement de la souris sur la grille pour dplacer la cellule slectionne.
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;

    // Ragit aux touches du clavier.
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;

    // Traite la commande du menu permettant de dfinir une cellule. 
    procedure DoValeurClick(Sender : TObject);
    procedure DoMarkClick(Sender : TObject);
  public
    constructor Create(Aowner : TComponent); override;
    destructor Destroy; override;

    property SudokuGrid : TSudokuBase read FSudokuGrid write SetSudokuGrid;
    property AffichageLettre : boolean read FAffichageLettre write SetAffichageLettre;
  end;

implementation
const AFFICHAGE : array[boolean] of string = (' 123456789ABCDEFG', ' ABCDEFGHIJKLMNOP');
{ TSudokuGrid }

constructor TGrilleSudoku.Create(Aowner: TComponent);
var
  i : integer;
  root : TMenuItem;
  item : TMenuItem;
begin
  inherited Create(AOwner);

  // Initialisation des proprits d'affichage de la grille.

  AffichageLettre := MODE_AFFICHAGE_LETTRE; //

  // Le nombre de lignes et colonnes de la grille est le mme que cellui d'une grille de Sudoku
  ColCount := LARGEUR;
  RowCount := HAUTEUR;

  // Pas de colonnes, ni lignes fixes.
  FixedCols := 0;
  FixedRows := 0;

  // On calcule les dimensions du composant en fonction de la taille voulue pour chaque cellule :
  Width := (CellWidth + CellThinBorder)*LARGEUR + // Largeur d'une cellule et de ses bordures
            CellStrongBorder*(LGBLOC-1); // Il faut ajouter la largeur des bordures des sous-blocs.

  Height := (CellHeight + CellThinBorder)*HAUTEUR + // Hauteur d'une cellule et de ses bordures
             CellStrongBorder*(HTBLOC-1); // Il faut ajouter la hauteur des bordures des sous-blocs.

  BorderStyle := bsNone;  // La grille n'a pas de bordures autre que celles qu'on dessine dans DrawCell. 
  DefaultDrawing := false;  // La grille sera compltement dessine par DrawCell.
  Options := [];  // Toutes options standard des grilles sont dsactives. 
  DoubleBuffered := true; // Pour viter tout clignotement.

  // A prsent, on dfinit les largeurs des colonnes.
  for i := 0 to ColCount -1 do
  begin
    // La largeur d'une colonne correspond normalement  la largeur d'une cellule et de sa bordure.
    ColWidths[i] := CellWidth + CellThinBorder;

    // Pour la dernire colonne d'un sous-bloc, la bordure est plus paisse. Donc on largit la colonne
    // d'autant.
    if ((i mod LGBLOC) = LGBLOC-1) and (i<>ColCount-1)
    then ColWidths[i] := ColWidths[i] + CellStrongBorder;
  end;

  // On dfinit les hauteurs des lignes
  for i := 0 to RowCount -1 do
  begin
    // La hauteur d'une ligne correspond normalement  la hautehr d'une cellule et de sa bordure.
    RowHeights[i] := CellHeight + CellThinBorder;;

    // Pour la dernire ligne d'un sous-bloc, la bordure est plus paisse. Donc on agrandit la ligne
    // d'autant.
    if ((i mod HTBLOC) = HTBLOC-1) and (i<>RowCount-1)
    then RowHeights[i] := RowHeights[i] + CellStrongBorder;
  end;

  // Enfin, on cre les objets qui serviront  dessiner les cellules de la grille.

  // Bordures.
  FBorderBrush := TBrush.Create;
  FBorderBrush.Style := bsSolid;
  FBorderBrush.Color := clBlack;

  // Fond d'une cellule normale.
  FCellBrush := TBrush.Create;
  FCellBrush.Style := bsSolid;
  FCellBrush.Color := clWhite;

  // Fond d'une cellule slectionne.
  FSelectedCellBrush := TBrush.Create;
  FSelectedCellBrush.Style := bsSolid;
  FSelectedCellBrush.Color := clSkyBlue;

  // Couleur d'une cellule rvle.
  FReveleeFont := TFont.Create;
  FReveleeFont.Name := 'Arial';
  FReveleeFont.Size := 20;
  FReveleeFont.Color := clBlack;

  // Couleur d'une cellule joue par le joueur (ou rsolue).
  FNormalFont := TFont.Create;
  FNormalFont.Name := 'Arial';
  FNormalFont.Size := 20;
  FNormalFont.Color := clGreen;

  // Couleur pour dessiner les candidats d'une cellule.
  FMarkFont := TFont.Create;
  FMarkFont.Name := 'Arial';
  FMarkFont.Size := 8;
  FMarkFont.Color := clBlack;

  FPopupMenu := TPopupMenu.Create(self);
  root := FPopupMenu.CreateMenuItem;
  root.Caption := 'Valeurs';
  FPopupMenu.Items.Add(root);
  for i := VALEUR_MIN to VALEUR_MAX do
  begin
    Item := FPopupMenu.CreateMenuItem;

    Item.Caption := AFFICHAGE[AffichageLettre][i+1];
    Item.OnClick := DoValeurClick;
    Item.Checked := false;
    Item.Tag := i;
    root.Add(item);
  end;

  FMarkRoot := FPopupMenu.CreateMenuItem;
  FMarkRoot.Caption := 'Candidats';
  FPopupMenu.Items.Add(FMarkRoot);
  for i := VALEUR_MIN to VALEUR_MAX do
  begin
    Item := FPopupMenu.CreateMenuItem;
    Item.Caption := AFFICHAGE[AffichageLettre][i+1];
    Item.OnClick := DoMarkClick;
    Item.Checked := false;
    Item.Tag := i;
    FMarkRoot.Add(item);
  end;  
end;

destructor TGrilleSudoku.Destroy;
begin
  FBorderBrush.Free;
  FCellBrush.Free;
  FSelectedCellBrush.Free;
  FReveleeFont.Free;
  FNormalFont.Free;
  FMarkFont.Free;
  FPopupMenu.Free;
  inherited Destroy;
end;

procedure TGrilleSudoku.DoMarkClick(Sender: TObject);
begin
  if SudokuGrid<>nil
  then begin
    SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, TMenuItem(Sender).Tag);
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
  end;
end;

procedure TGrilleSudoku.DoValeurClick(Sender: TObject);
begin
  if SudokuGrid<>nil
  then begin
    SudokuGrid.DefinirCellule(FCurrentCell.X, FCurrentCell.Y, TMenuItem(Sender).Tag);
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
  end;
end;

procedure TGrilleSudoku.DrawCell(ACol, ARow: Integer; ARect: TRect; AState: TGridDrawState);
var
  Rect: TRect;
  s : string;
  w : integer;
  h : integer;
  x, y: integer;
  i : integer;
  n : integer;
  ValeurCellule : cardinal;
  IsRevelee : boolean;
begin
  // On commence par dessiner les bordures
  Canvas.Brush := FBorderBrush;
  Canvas.FillRect(ARect);

  Rect := ARect;
  Rect.Left := Rect.Left + 1;
  Rect.Top := Rect.Top +1;
  Rect.Right := Rect.Left + CellWidth -1;
  Rect.Bottom := Rect.Top + CellHeight -1;
  if (ACol = FCurrentCell.X) and (ARow = FCurrentCell.Y)
  then Canvas.Brush := FSelectedCellBrush 
  else Canvas.Brush := FCellBrush;
  
  Canvas.FillRect(Rect);

  if SudokuGrid<>nil
  then begin
    ValeurCellule := SudokuGrid.ValeurCellule(ACol, ARow);
    IsRevelee := SudokuGrid.IsRevelee(ACol, ARow);

    if ValeurCellule<>CELLULE_VIDE 
    then begin
      if IsRevelee
      then Canvas.Font := FReveleeFont
      else Canvas.Font := FNormalFont;

      s := AFFICHAGE[AffichageLettre][ValeurCellule+1];
      w := Canvas.TextWidth(s);
      h := Canvas.TextHeight(s);
      Canvas.TextRect(Rect,
        Rect.Left + (Rect.Right-Rect.Left-w) div 2,
        Rect.Top + (Rect.Bottom-Rect.Top-h) div 2, s);
    end;

    // A prsent, on dessine les marks des cellules.
    Canvas.Font := FMarkFont;
    x := Rect.Left + 2;
    y := Rect.Bottom - Canvas.TextHeight('9');
    n := 0;
    for i := VALEUR_MIN to VALEUR_MAX do
    begin
      if SudokuGrid.IsCandidat(ACol, ARow, i)
      then begin
        s := AFFICHAGE[AffichageLettre][i+1];      
        Canvas.TextOut(x, y, s);
        x := x + Canvas.TextWidth(s);
        inc(n);
        if (n = VALEUR_MAX div 2+1)
        then begin
          x := Rect.Left+2;
          y := ARect.Top+1;
        end;
      end;
    end;
  end;
end;

procedure TGrilleSudoku.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);

  if Shift = []
  then begin
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
    case Key of
    VK_LEFT:
      if FCurrentCell.X > 0
      then FCurrentCell.X := FCurrentCell.X-1;
    VK_RIGHT:
      if FCurrentCell.X <8
      then FCurrentCell.X := FCurrentCell.X+1;
    VK_UP:
      if FCurrentCell.Y > 0
      then FCurrentCell.Y := FCurrentCell.Y-1;
    VK_DOWN:
      if FCurrentCell.Y < 8
      then FCurrentCell.Y := FCurrentCell.Y+1;
    end;
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
  end;

  if Shift = [ssCtrl]
  then begin
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
    case Key of
    VK_NUMPAD0:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 0);
    VK_NUMPAD1:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 1);
    VK_NUMPAD2:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 2);
    VK_NUMPAD3:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 3);
    VK_NUMPAD4:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 4);
    VK_NUMPAD5:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 5);
    VK_NUMPAD6:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 6);
    VK_NUMPAD7:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 7);
    VK_NUMPAD8:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 8);
    VK_NUMPAD9:
      SudokuGrid.MarkerCellule(FCurrentCell.X, FCurrentCell.Y, 9);
    end;
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);  
  end;
end;

procedure TGrilleSudoku.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);

  if (SudokuGrid<>nil)
  then begin
    if (Key>='0') and (Key<='9')
    then begin
      SudokuGrid.DefinirCellule(FCurrentCell.X, FCurrentCell.Y, StrToInt(Key));
      SudokuGrid.RechercherTousLesCandidats;
      Invalidate;
      InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
    end;
  end;
end;

procedure TGrilleSudoku.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  pt : TPoint;
  i : integer;
begin
  inherited MouseDown(Button, Shift, X, Y);

  if Button = mbRight
  then begin
    FCurrentCell := MouseCoord(X, y);
    pt := ClientToScreen(Point(X, Y));
    if SudokuGrid<>nil
    then begin
      for i := 0 to FMarkRoot.Count-1 do
      begin
        FMarkRoot.Items[i].Checked := SudokuGrid.IsCandidat(FCurrentCell.X, FCurrentCell.Y, i+1);
      end;
    end;
    FPopupMenu.Popup(pt.X, pt.y);
  end;
end;

procedure TGrilleSudoku.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  r : TGridCoord;
begin
  r := MouseCoord(X, Y);
  inherited MouseMove(Shift, X, Y);
  if (r.X<>FCurrentCell.X) or (r.Y<>FCurrentCell.Y)
  then begin
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
    FCurrentCell := r;
    InvalidateCell(FCurrentCell.X, FCurrentCell.Y);
  end;
end;

procedure TGrilleSudoku.SetAffichageLettre(const Value: boolean);
begin
  FAffichageLettre := Value;
  Invalidate;
end;

procedure TGrilleSudoku.SetSudokuGrid(const Value: TSudokuBase);
begin
  if FSudokuGrid<>Value
  then begin
    FSudokuGrid := Value;
    Invalidate;
  end;
end;

end.
