unit uSudokuMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, uSudokuGrid, Menus, Grids, ImgList, ToolWin, ComCtrls, ActnList,
  uSudokuBase;

type
  TSudokuMain = class(TForm)
    MainMenu1: TMainMenu;
    Fichier1: TMenuItem;
    Rsolutioncomplte1: TMenuItem;
    Nouveau1: TMenuItem;
    Grillevide1: TMenuItem;
    iescargot1: TMenuItem;
    Grilleminimalede1: TMenuItem;
    Jouerlagrille1: TMenuItem;
    grdSolutions: TDrawGrid;
    N1: TMenuItem;
    Quiter1: TMenuItem;
    tbMain: TToolBar;
    ImageList2: TImageList;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    acLst: TActionList;
    acNouveau: TAction;
    acResoudre1: TAction;
    acOpen: TAction;
    acSave: TAction;
    Rsoudrelagrille1: TMenuItem;
    Enregistrer1: TMenuItem;
    N2: TMenuItem;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    opDlg: TOpenDialog;
    svDlg: TSaveDialog;
    Rsolutionnonoptimise1: TMenuItem;
    acRechercherCoupsPossibles: TAction;
    Rechercherlescoupspossibles1: TMenuItem;
    Grillevide2: TMenuItem;
    acAbout: TAction;
    Apropos1: TMenuItem;
    Apropos2: TMenuItem;
    procedure acRechercherCoupsPossiblesExecute(Sender: TObject);
    procedure acSaveExecute(Sender: TObject);
    procedure acOpenExecute(Sender: TObject);
    procedure acResoudre1Execute(Sender: TObject);
    procedure acNouveauExecute(Sender: TObject);
    procedure Quiter1Click(Sender: TObject);
    procedure grdSolutionsDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState);
    procedure grdSolutionsSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
    procedure Jouerlagrille1Click(Sender: TObject);
    procedure Grilleminimalede1Click(Sender: TObject);
    procedure iescargot1Click(Sender: TObject);
    procedure Grillevide1Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Rsolutioncomplte1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure acResoudre1Update(Sender: TObject);
    procedure acAboutExecute(Sender: TObject);
  private
    FGrillesTest : TStrings;    // Liste des grilles de Gordon Royle
    procedure acSudokuExecute(Sender: TObject);

    // Charge la grille dfinie par GridStr.
    procedure NouvelleGrille(GridStr : string);

    // Affiche la solution no <no>. <no> dsigne une des solution prsentes dans Grille.ListeSolution
    procedure AfficherSolution(no : integer);

  public
    DebutResolution : int64;    // Date de lancemant de la rsolution des grilles.
    Frequence : int64;          // frquence du compteur de performance.

    Grille : TSudokuBase; // Grille de Sudoku en cours.
    Grid : TGrilleSudoku;     // Composant affichant la grille de sudoku.
  end;

var
  SudokuMain: TSudokuMain;

implementation
uses uFrmGordonRoyle, ufrmAbout;


{$R *.dfm}

const
  GrilleVide = '000000000000000000000000000000000000000000000000000000000000000000000000000000000';
  GrilleAIEscargot = '100007090030020008009600500005300900010080002600004000300000010040000007007000300';

  NBMAXSOLUTION = 1000000; // On arrte la recherche des solutions lorsqu'on a trouv NBMAXSOLUTION solutions
//  NBMAXSOLUTION = 1;

procedure TSudokuMain.acAboutExecute(Sender: TObject);
var
  dlg : TForm;
begin
  dlg := TfrmAbout.Create(self);
  try
    dlg.ShowModal;
  finally
    dlg.Free;
  end;
end;

procedure TSudokuMain.acNouveauExecute(Sender: TObject);
begin
  NouvelleGrille(GrilleVide);
  Caption := 'Grille vide';
end;

procedure TSudokuMain.acOpenExecute(Sender: TObject);
begin
  if opDlg.Execute
  then begin
    Grille.LoadFromFile(opDlg.FileName);
    grid.Invalidate;
  end;
end;

procedure TSudokuMain.acRechercherCoupsPossiblesExecute(Sender: TObject);
begin
  Grille.RechercherTousLesCandidats;
  Grid.Invalidate;
end;

procedure TSudokuMain.acResoudre1Execute(Sender: TObject);
var
  i, j : integer;
  nbRevelee : integer;
  nbMax : integer;
  
begin
  // Tout d'abord, on vrifie si l'utilisateur n'a pas oublier de lancer la commande Jouer la grille.
  nbRevelee := 0;
  for i := 0 to LARGEUR-1 do
  begin
    for j := 0 to HAUTEUR-1 do
    begin
      if Grille.IsRevelee(i, j)
      then inc(nbRevelee);
    end;
  end;

  if (nbRevelee=0)
  then begin
    Grille.JouerLaGrille;
  end;

  // A prsent, on peut prparer la rsolution de la grille.
  Screen.Cursor := crHourGlass;
  nbMax := NBMAXSOLUTION;
  grid.Invalidate;

  // Rinitialise la grille pour recommencer  repartir de sa dfinition "vide".  
  Grille.Initialiser;
  Grille.ListeSolution.Clear;

  // A prsent, on va pouvoir lancer la rsolution. On peut lancer le chrono.
  QueryPerformanceCounter(DebutResolution);

  // On commence par initialisation la rsolution de la grille.
  Grille.InitialiserResolution;
//  QueryPerformanceCounter(DebutResolution);
  Grille.Resoudre(nbMax); // puis on peut lancer la rsolution relle.

  // A prsent, on lit la frquence du chrono pour pouvoir convertir les dures en ms.
  QueryPerformanceFrequency(Frequence);

  // Enfin, on initialise l'affichage des solutions.
  grdSolutions.RowCount := Grille.ListeSolution.Count +2;
  if Grille.ListeSolution.Count> 0
  then AfficherSolution(0);

  grdSolutions.Invalidate;
  Screen.Cursor := crDefault;
end;

procedure TSudokuMain.acResoudre1Update(Sender: TObject);
begin
  TAction(Sender).Caption := 'Rsoudre : ' + Grille.GetAlgorithmName;
end;

procedure TSudokuMain.acSaveExecute(Sender: TObject);
begin
  if svDlg.Execute
  then begin
    Grille.SaveToFile(svDlg.Filename);
  end;
end;

procedure TSudokuMain.acSudokuExecute(Sender: TObject);
var
  SudokuClass : TSudokuClass;
  Grille2 : TSudokuBase;
  i,j : integer;

begin
  SudokuClass := TSudokuClass(FindClass(ListeAlgorithme[TAction(Sender).Tag]));

  // On cre une nouvelle instance de la grille, utilisant la classe dfinie
  // pour l'algorithme de rsolution choisit. 
  Grille2 := SudokuClass.Create;

  // Ensuite, on recopie les valeurs de la grille
  for i := 0 to HAUTEUR-1 do
  begin
    for j := 0 to LARGEUR-1 do
    begin
      if Grille.IsRevelee(i, j)
      then Grille2.DefinirCellule(i, j, Grille.ValeurCellule(i, j))
      else Grille2.DefinirCellule(i, j, CELLULE_VIDE);
    end;
  end;
  Grille2.JouerLaGrille; // Marque toutes les cellules remplies comme tant des rvls.

  // La nouvelle grille est prte. Il ne reste plus qu' la substituer  la grille
  // actuelle.
  Grille.Free;
  Grille := Grille2;

  Grid.SudokuGrid := Grille; // La vue de la grille doit afficher la nouvelle instance.
  Grid.Invalidate;

  // On met  jour l'affichage de la liste des solutions trouves.
  grdSolutions.RowCount := Grille.ListeSolution.Count +2;
  grdSolutions.Invalidate;

  // Enfin, on peut lancer la rsolution de la grille :
  acResoudre1.Execute;
end;

procedure TSudokuMain.AfficherSolution(no: integer);
var
  s : string;
  n : integer;
  x, y : integer;
begin
  s := Grille.ListeSolution[no];

  // D'abord, on efface la grille.
  Grille.Initialiser;

  // Ensuite, on dfinit la solution trouve.
  n := 1;
  for y := 0 to HAUTEUR-1 do
  begin
    for x := 0 to LARGEUR-1 do
    begin
      Grille.DefinirCellule(x, y, ord(s[n])-65);
      inc(n);
    end;
  end;
  Grid.Invalidate; // Il ne reste plus qu'a redessiner la grille.
end;

procedure TSudokuMain.FormCreate(Sender: TObject);
var
  filename : string;
  i : integer;
  Action : TAction;
  SudokuClass : TSudokuClass;
  menuItem : TMenuItem;
begin
  // Initialisation dynamique de la forme :
  // On commence par crer dynamiquement le composant d'affichage de la grille de jeu.
  Grid := TGrilleSudoku.Create(self);
  Grid.Parent := Self;
  Grid.Left := 0;
  grid.Top := tbMain.Height;

  // Ensuite, on ajuste les dimensions de la form.
  Width := grid.Width + Width - ClientWidth + grdSolutions.Width;
  Height := grid.Height + Height - ClientHeight + tbMain.Height;

  // Enfin on charge les grilles de Gordon Royle.
  FGrillesTest := TStringList.Create;
  filename := GetCurrentDir + '\sudoku17.txt';
  if FileExists(filename)
  then FGrillesTest.LoadFromFile(filename)
  else begin
    Grilleminimalede1.Enabled := false;
    ShowMessage('Impossible de trouver le fichier ' + QuotedStr(filename) + #13#10 +
                'Les grilles de Gordon Royle seront dsactives !');
  end;

  FGrillesTest.Insert(0, GrilleVide);  // On insre une grille vide  la position 0

  Randomize;

  // Par dfaut, on affiche une grille vide.
  NouvelleGrille(GrilleVide);

  // Initialisation de la liste des mthodes de rsolution.
  for i := 0 to ListeAlgorithme.Count -1 do
  begin
    Action := TAction.Create(acLst);
    Action.Name := 'acResoudre' + IntToStr(i+2);
    Action.ActionList := acLst;
    SudokuClass := TSudokuClass(FindClass(ListeAlgorithme[i]));
    Action.Caption := SudokuClass.GetAlgorithmName;
    Action.Hint := 'Rsoudre la grille en utilisant : ' + SudokuClass.GetAlgorithmName;
    Action.Tag := i; // On mmorise le numro de la mthode  utiliser
    Action.OnExecute := acSudokuExecute;

    menuItem := TMenuItem.Create(Rsolutioncomplte1);
    menuItem.Name := 'mnResoudre' + IntToStr(i+2);
    menuItem.Action := Action;
    Rsolutioncomplte1.Add(menuItem);
  end;
end;

procedure TSudokuMain.FormDestroy(Sender: TObject);
begin
  FGrillesTest.Free;
end;

procedure TSudokuMain.grdSolutionsDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
  State: TGridDrawState);
var
  FinStr : string;
  Fin : int64;
  procedure DrawCenter(s: string);
  var w : integer;
  begin
    w := grdSolutions.Canvas.TextWidth(s);
    grdSolutions.Canvas.TextRect(Rect, Rect.Left + (Rect.Right - Rect.Left - w) div 2, Rect.Top +2, s);
  end;

  procedure DrawRight(s : string);
  var w : integer;
  begin
    w := grdSolutions.Canvas.TextWidth(s);
    grdSolutions.Canvas.TextRect(Rect, Rect.Right - w - 2, Rect.Top +2, s);
  end;
begin
  if ARow = 0
  then begin
    case ACol of
    0 : DrawCenter('N');
    1 : DrawCenter('Dure');
    end;
  end
  else begin
    if ARow-1 < Grille.ListeSolution.Count
    then begin
      case ACol of
      0 : DrawRight(IntToStr(ARow));
      1 : begin
            FinStr := copy(Grille.ListeSolution[ARow-1], LARGEUR*HAUTEUR+4, 16);
            Fin := StrToInt64(FinStr);
            DrawRight(Format('%0.4f ms', [(fin-DebutResolution)*1000 / Frequence]));
          end;
      end;
    end
    else DrawRight('');
  end;
end;

procedure TSudokuMain.grdSolutionsSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
begin
  CanSelect := true;
  if (ARow > 0) and (ARow-1 < Grille.ListeSolution.Count)
  then AfficherSolution(ARow-1);
end;

procedure TSudokuMain.Grilleminimalede1Click(Sender: TObject);
var
  dlg : TfrmGordonRoyle;
begin
  dlg := TfrmGordonRoyle.Create(self);
  try
    if dlg.ShowModal = mrOk
    then begin
      NouvelleGrille(FGrillesTest[dlg.GetGridNo]);
      Caption := 'Grille n' + IntToStr(dlg.GetGridNo) + ' de Gordon Royle, The University of Western Australia';
    end;
  finally
    dlg.Free;
  end;
end;

procedure TSudokuMain.Grillevide1Click(Sender: TObject);
begin
  NouvelleGrille(GrilleVide);
  Caption := 'Grille vide';
end;

procedure TSudokuMain.iescargot1Click(Sender: TObject);
begin
  NouvelleGrille(GrilleAIEscargot);
  Caption := 'Grille AI-Escargot, de Arto Inkala';
end;

procedure TSudokuMain.Jouerlagrille1Click(Sender: TObject);
begin
  Grille.JouerLaGrille;
  Grid.Invalidate;
end;

procedure TSudokuMain.NouvelleGrille(GridStr: string);
var
  i,j : integer;
  n : integer;
  Valeur : integer;
  SudokuClass : TSudokuClass;
begin
  if Assigned(Grille)
  then begin
    SudokuClass := TSudokuClass(Grille.ClassType);
    Grille.Free;
  end
  else SudokuClass := TSudokuClass(FindClass(ListeAlgorithme[ListeAlgorithme.Count-1]));

  // Puis on cre une nouvelle grille vide.
  Grille := SudokuClass.Create;

  // Enfin, on dfinit les cellules rvles en fonction de la dfinition fournie par GridStr.
  n := 1;
  for i := 0 to HAUTEUR-1 do
  begin
    for j := 0 to LARGEUR-1 do
    begin
      if n<= length(GridStr)
      then Valeur := StrToIntDef(GridStr[n], CELLULE_VIDE)
      else Valeur := CELLULE_VIDE;
      Grille.DefinirCellule(j, i, Valeur);
      inc(n);
    end;
  end;
  Grille.JouerLaGrille;

  // Enfin, on prpare la mise  jour de l'affichage.

  Grid.SudokuGrid := Grille; // La vue de la grille doit afficher la nouvelle instance.
  Grid.Invalidate;

  // On met  jour l'affichage de la liste des solutions trouves.
  grdSolutions.RowCount := Grille.ListeSolution.Count +2;
  grdSolutions.Invalidate;
end;

procedure TSudokuMain.Quiter1Click(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TSudokuMain.Rsolutioncomplte1Click(Sender: TObject);
var
  nbMax : integer;

begin
  Screen.Cursor := crHourGlass;
  nbMax := NBMAXSOLUTION;
  grid.Invalidate;
  Grille.Initialiser;
  Grille.ListeSolution.Clear;

  QueryPerformanceCounter(DebutResolution);
  Grille.InitialiserResolution;
  Grille.Resoudre(nbMax);
  QueryPerformanceFrequency(Frequence);

  grdSolutions.RowCount := Grille.ListeSolution.Count +2;
  if Grille.ListeSolution.Count> 0
  then AfficherSolution(0);
  grdSolutions.Invalidate;
  Screen.Cursor := crDefault;
end;

end.
