unit uSudokuNaifBase;

interface
uses classes, sysutils, windows, uSudokuBase;

type
  TCellule = record
    Valeur : integer;     // Valeur de la cellule
    IsRevelee : boolean; // indique si la cellule est revlle : Sa valeur est fournie au dpart.
    FlgCandidats : array[VALEUR_MIN..VALEUR_MAX] of boolean; // dsigne les marques dfinies par le joueur
  end;

  PCellule = ^TCellule;

  TSudokuNaifBase = class(TSudokuBase)
  protected
    FOwnSolution : boolean;
    // Dfinition de la grille de Sudoku
    FGrille : array[0..LARGEUR-1, 0..HAUTEUR-1] of TCellule;

    // Retourne le contenu de la grille sous la forme d'un chaine de LARGEUR*HAUTEUR caractres, avec
    // A = 0, B = 1 ...
    function DumpGrid : string;

    // Ajoute la grille actuelle  la liste des solutions trouves.
    procedure MemoriserSolution;

    // Dfinit la valeur de la cellule <X,Y>  <Valeur>.
    // Cette mthode ne vrifie pas si le coup est valide.
    procedure DefinirValeur(X, Y, Valeur : integer); override;

    // Dfinit la cellule <X,Y> comme tant une celule rvle.
    procedure DefinirRevelee(X, Y : integer); override;


  public
    // Cre et initialise une nouvelle instance de la classe TGrilleSudoku.
    // La grille est initialise sur une grille vide.
    constructor Create; override;

    destructor Destroy; override;

    // Efface le contenu de la grille pour qu'on puisse lui dfinir une nouvelle
    // grille.
    procedure EffacerGrille; override;

    // Ajoute/supprime un marqueur de candidat possible pour la cellule <x, y>.
    // Cette mthode sert au joueur humain qui tente de rsoudre la grille.
    procedure MarkerCellule(x, y : integer; candidat : integer); override;

    // Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
    // modifies et doivent tre dessine en noir.
    function IsRevelee(x, y : integer) : boolean; override;

    // Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
    function ValeurCellule(X, Y : integer) : integer; override;

    // Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
    function IsCandidat(X, Y, Valeur : integer) : boolean; override;
  end;

implementation

// Cre et initialise une nouvelle instance de la classe TSudoku.
// La grille est initialise sur une grille vide.
constructor TSudokuNaifBase.Create;
begin
  inherited Create;
  FillChar(FGrille, sizeof(FGrille), CELLULE_VIDE);
  FOwnSolution := true;
end;

// Dfinit la cellule <X,Y> comme tant une celule rvle.
procedure TSudokuNaifBase.DefinirRevelee(X, Y: integer);
begin
  FGrille[x, y].IsRevelee := true;
end;

// Dfinit la valeur de la cellule <X,Y>  <Valeur>.
// Cette mthode ne vrifie pas si le coup est valide.
procedure TSudokuNaifBase.DefinirValeur(X, Y, Valeur: integer);
begin
   FGrille[x, y].Valeur := valeur;
end;

destructor TSudokuNaifBase.Destroy;
begin
  if not FOwnSolution
  then ListeSolution := nil;

  inherited Destroy;
end;

// Retourne le contenu de la grille sous la forme d'un chaine de LARGEUR*HAUTEUR caractres, avec
// A = 0, B = 1 ...
function TSudokuNaifBase.DumpGrid: string;
var
  i,j : integer;
  n : integer;
begin
  n := 1;
  SetLength(result, LARGEUR*HAUTEUR);
  for j := 0 to HAUTEUR-1 do
  begin
    for i := 0 to LARGEUR-1 do
    begin
      result[n] := chr(FGrille[i,j].Valeur + 65);
      inc(n);
    end;
  end;
end;

procedure TSudokuNaifBase.EffacerGrille;
begin
  Fillchar(FGrille, sizeof(FGrille), 0);
end;

// Indique si la valeur <Valeur> est dsigne comme tant un candidat pour la cellule <X,Y>.
function TSudokuNaifBase.IsCandidat(X, Y, Valeur: integer): boolean;
begin
  result := FGrille[x, y].FlgCandidats[Valeur];
end;

// Indique si la cellule <X, Y> est une cellule rvle. Les cellules rvle ne peuvent pas tre
// modifies et doivent tre dessine en noir.
function TSudokuNaifBase.IsRevelee(x, y: integer): boolean;
begin
  result := FGrille[x, y].IsRevelee;
end;

// Ajoute/supprime un marqueur de valeur possible pour la cellule <x, y>.
// Cette mthode sert au joueur humain qui tente de rsoudre la grille.
procedure TSudokuNaifBase.MarkerCellule(x, y, candidat: integer);
begin
  if not FGrille[x, y].IsRevelee
  then FGrille[x,y].FlgCandidats[candidat] := not FGrille[x,y].FlgCandidats[candidat];
end;

// Ajoute la grille actuelle  la liste des solutions trouves.
procedure TSudokuNaifBase.MemoriserSolution;
var
  t0 : int64;
  s : string;
begin
  // Tout d'abord, on mmorise la date  laquelle la solution a t trouve.
  QueryPerformanceCounter(t0);

  s := DumpGrid;
  s := s + ' - ' + IntToStr(t0); // On ajoute la date de rsolution  la grille trouve.
  ListeSolution.Add(s);
end;

// Retourne la valeur de la cellule <X,Y>. Les cellules vides retournent CELLULE_VIDE.
function TSudokuNaifBase.ValeurCellule(X, Y: integer): integer;
begin
  result := FGrille[x, y].Valeur;
end;

end.
